import {
  Body,
  Controller,
  Get,
  HttpException,
  HttpStatus,
  Patch,
  Post,
  Query,
  Redirect,
  Req,
  UseGuards,
} from '@nestjs/common';
import { ApiBody, ApiQuery, ApiResponse } from '@nestjs/swagger';
import { ConfigService } from '@nestjs/config';
import { AuthService } from './auth.service';
import { JwtGuard } from './guards/jwt.guard';
import { RadarScopeDto } from './dto/radarscope.dto';
import { RadarTokenDto } from './dto/radartoken.dto';
import { UserDto } from './dto/user.dto';
import { Request } from 'express';

@Controller('auth')
export class AuthController {
  constructor(
    private config: ConfigService,
    private authService: AuthService,
  ) {}

  @Get('/vatsim')
  @ApiQuery({
    name: 'code',
    description: 'The authorization code',
    type: String,
  })
  @Redirect()
  async vatsim(@Query('code') code) {
    if (!code) {
      throw new HttpException(
        'Code not found in the query',
        HttpStatus.NOT_ACCEPTABLE,
      );
    }

    const token = await this.authService.login(code);

    if (token !== undefined) {
      return {
        url: `${this.config.get<string>(
          'frontend.base-url',
        )}/${this.config.get<string>(
          'frontend.login-endpoint',
        )}?token=${token}`,
      };
    } else {
      return {
        url: `${this.config.get<string>(
          'frontend.base-url',
        )}/${this.config.get<string>('frontend.login-endpoint')}`,
      };
    }
  }

  @Post('/radarScope')
  @ApiBody({
    description: 'The airport definition',
    type: RadarScopeDto,
  })
  @ApiResponse({
    status: 200,
    description: 'The created Bearer token to use endpoints',
    type: RadarTokenDto,
  })
  @ApiResponse({
    status: 404,
    description: 'The VATSIM ID and key combination is invalid',
  })
  async radarScope(
    @Body('scopeData') scopeData: RadarScopeDto,
  ): Promise<RadarTokenDto> {
    return this.authService
      .loginRadarScope(scopeData.vatsimId, scopeData.key)
      .then((token) => {
        if (token === undefined) {
          throw new HttpException(
            'Unknown VATSIM ID or invalid key',
            HttpStatus.NOT_FOUND,
          );
        }
        return { token };
      });
  }

  @UseGuards(JwtGuard)
  @Get('/validate')
  @ApiResponse({
    status: 200,
    description: 'The created Bearer token to use endpoints',
  })
  @ApiResponse({
    status: 401,
    description: 'The sent Bearer token is invalid',
  })
  // eslint-disable-next-line @typescript-eslint/no-empty-function
  validate(): void {}

  @UseGuards(JwtGuard)
  @Get('/user')
  async user(@Req() request: Request): Promise<UserDto> {
    const token = request.headers.authorization.replace('Bearer ', '');
    return this.authService.user(token).then((user) => {
      if (!user) {
        throw new HttpException('Invalid user request', HttpStatus.NOT_FOUND);
      }

      return {
        vatsimId: user.vatsimId,
        fullName: user.fullName,
        radarScopeKey: user.radarScopeKey,
        administrator: user.administrator,
        airportConfigurationAccess: user.airportConfigurationAccess,
      };
    });
  }

  @UseGuards(JwtGuard)
  @Patch('/refreshRadarScopeKey')
  async refreshRadarScopeKey(@Req() request: Request): Promise<void> {
    const token = request.headers.authorization.replace('Bearer ', '');
    return this.authService.resetRadarScopeKey(token);
  }
}
